"""
These functions are portable.

read: only reading.
build: put parts in to a whole.
calc: transforming or reform.
is: boolean functions.
"""

import json
import secrets
from typing import List, Union

from thor_devkit import abi, cry, transaction
from thor_devkit.cry import address, secp256k1

from .wallet import Wallet


def build_url(base: str, tail: str) -> str:
    """ Build a proper URL, base + tail """
    return base.rstrip("/") + "/" + tail.lstrip("/")


def read_json_file(path_like: str) -> dict:
    """ Read json file """
    with open(path_like, "r") as f:
        return json.load(f)


def build_params(types: List, args: List) -> bytes:
    """ ABI encode params according to types """
    return abi.Coder.encode_list(types, args)


def calc_address(priv: bytes) -> str:
    """ Calculate an address from a given private key """
    public_key = secp256k1.derive_publicKey(priv)
    _address_bytes = cry.public_key_to_address(public_key)
    address = "0x" + _address_bytes.hex()
    return address


def calc_nonce() -> int:
    """ Calculate a random number for nonce """
    return int(secrets.token_hex(8), 16)


def calc_blockRef(block_id: str) -> str:
    """ Calculate a blockRef from a given block_id, id should starts with 0x """
    if not block_id.startswith("0x"):
        raise Exception("block_id should start with 0x")
    return block_id[0:18]


def calc_chaintag(hex_str: str) -> int:
    """ hex_str can be both like '0x4a' or just '4a' """
    return int(hex_str, 16)


def calc_gas(vm_gas: int, intrinsic_gas: int) -> int:
    """ Calculate recommended (safe) gas """
    return vm_gas + intrinsic_gas + 15000


def calc_vtho(gas: int, coef: 0) -> int:
    """ Calculate extimated vtho from gas """
    if coef > 255 or coef < 0:
        raise Exception("coef: [0~255]")
    return gas * (1 + coef / 255)


def any_emulate_failed(emulate_responses: List) -> bool:
    """ Check the emulate response, if any tx reverted then it is a fail """
    results = [each["reverted"] for each in emulate_responses]
    return any(results)


def read_vm_gases(emulated_responses: List) -> List[int]:
    """ Extract vm gases from a batch of emulated executions. """
    results = [int(each["gasUsed"]) for each in emulated_responses]
    return results


def build_tx_body(
    clauses: List,
    chainTag: int,
    blockRef: str,
    nonce: int,
    expiration: int = 32,
    gasPriceCoef: int = 0,
    gas: int = 0,
    dependsOn=None,
) -> dict:
    """Build a Tx body.

    Clause should confine to "thor_devkit.transaction.CLAUSE" schema. {to, value, data}
    Tx body shall confine to "thor_devkit.transaction.BODY" schema.
    """
    body = {
        "chainTag": chainTag,
        "blockRef": blockRef,
        "expiration": expiration,
        "clauses": clauses,
        "gasPriceCoef": gasPriceCoef,
        "gas": str(gas),
        "dependsOn": dependsOn,
        "nonce": nonce,
    }

    # Raise Exception if format check cannot pass.
    transaction.Transaction(body)

    return body


def calc_emulate_tx_body(caller: str, tx_body: dict) -> dict:
    """ Rip an emulated tx body from a normal tx body."""
    # Raise Exception if format check cannot pass.
    transaction.Transaction(tx_body)
    # Check if caller is correct format.
    if not address.is_address(caller):
        raise Exception(f"{caller} is not an address")

    e_tx_body = {
        "caller": caller,
        "blockRef": tx_body["blockRef"],
        "expiration": tx_body["expiration"],
        "clauses": tx_body["clauses"],
    }

    # Set gas field only when the tx_body set it.
    if int(tx_body["gas"]) > 0:
        e_tx_body["gas"] = tx_body["gas"]

    return e_tx_body


def calc_tx_unsigned(
    tx_body: dict, encode=False
) -> Union[transaction.Transaction, str]:
    """ Build unsigned transaction from tx body """
    tx = transaction.Transaction(tx_body)
    if encode:
        return "0x" + tx.encode().hex()
    else:
        return tx


def calc_tx_signed(
    wallet: Wallet, tx_body: dict, encode=False
) -> Union[transaction.Transaction, str]:
    """ Build signed transaction from tx body """
    tx = calc_tx_unsigned(tx_body, encode=False)
    message_hash = tx.get_signing_hash()
    signature = wallet.sign(message_hash)
    tx.set_signature(signature)
    if encode:
        return "0x" + tx.encode().hex()
    else:
        return tx


def is_readonly(abi_dict: dict):
    """ Check the abi, see if the function is read-only """
    # Check the shape of input data
    abi.FUNCTION(abi_dict)

    return (
        abi_dict["stateMutability"] == "view" or abi_dict["stateMutability"] == "pure"
    )


def calc_revertReason(data: str) -> str:
    """
    Extract revert reason from data

    Parameters
    ----------
    data : str
        "0x...." string of data

    Returns
    -------
    str
        revert reason in string

    Raises
    ------
    Exception
        If the revert reason data is not in good shape
    """
    if not data.startswith("0x08c379a0"):  # abi encoded of "Error(string)"
        raise Exception(f"{data.hex()} is not a valid error message")

    message = abi.Coder.decode_single(
        "string", bytes.fromhex(data.replace("0x08c379a0", ""))
    )
    return message


def is_reverted(receipt: dict) -> bool:
    """ Check receipt to see if tx is reverted """
    return receipt["reverted"]


def read_created_contracts(receipt: dict) -> list:
    """ Read receipt and return a list of contract addresses created """
    a = []
    for x in receipt["outputs"]:
        if x.get("contractAddress"):
            a.append(x.get("contractAddress"))

    return a


def is_contract(account: dict) -> bool:
    """ Check if the address online is a contract """
    return account["hasCode"]



from thor_devkit import cry
from thor_devkit.cry import mnemonic, keystore, secp256k1


def check(priv: bytes):
    if len(priv) != 32:
        raise Exception("Private key should be 32 bytes")


class Wallet:
    def __init__(self, priv: bytes):
        check(priv)
        self.priv: bytes = priv
        self.public: bytes = secp256k1.derive_publicKey(self.priv)
        self.address_bytes: bytes = cry.public_key_to_address(self.public)
        self.address: str = "0x" + self.address_bytes.hex()

    @classmethod
    def fromPrivateKey(cls, priv: bytes):
        """ Get a wallet from private key """
        return cls(priv)

    @classmethod
    def fromMnemonic(cls, words: list):
        """ Get a wallet from words """
        if not mnemonic.validate(words):
            raise Exception("Words validation failed, check spelling?")
        priv = mnemonic.derive_private_key(words)
        return cls(priv)

    @classmethod
    def fromKeyStore(cls, ks: dict, password: bytes):
        """ Get a wallet from keystore (dict) """
        if not keystore.well_formed(ks):
            raise Exception("Keystore is not well formed")

        priv = keystore.decrypt(ks, password)
        return cls(priv)

    def sign(self, msg_hash: bytes) -> bytes:
        """ Sign the stuff with this wallet """
        signature = secp256k1.sign(msg_hash, self.priv)
        return signature

    def verifySignature(self, msg_hash: bytes, signature: bytes) -> bool:
        """ Verify signature if signed by us. """
        public_key = secp256k1.recover(msg_hash, signature)
        return public_key == self.public

    def getAddress(self) -> str:
        return self.address



""" Contract is a representation of an underlying Solidity compiled JSON """
from typing import Union, List
from .utils import read_json_file
from thor_devkit import abi


class Contract:
    def __init__(self, meta_dict: dict):
        self.contract_meta: dict = meta_dict

    @classmethod
    def fromFile(cls, path_or_str):
        meta_dict = read_json_file(path_or_str)
        return cls(meta_dict)

    def get_contract_name(self) -> Union[str, None]:
        """ Get the smart contract name, or None """
        return self.contract_meta.get("contractName")

    def get_bytecode(self, key: str = "bytecode") -> bytes:
        """ Get bytecode of this smart contract """
        return bytes.fromhex(self.contract_meta[key])

    def get_abis(self) -> List[dict]:
        """ Get ABIs of this contract as a list """
        return self.contract_meta["abi"]

    def get_abi(self, func_name: str) -> Union[dict, None]:
        """ Get specific ABI by function/event name, or None if not found """
        abis = self.get_abis()
        targets = [each for each in abis if each.get("name") == func_name]
        assert len(targets) <= 1  # Zero or at most, one found.
        if len(targets):
            return targets[0]
        else:
            return None

    def get_function_by_name(self, func_name: str) -> Union[abi.Function, None]:
        """ Get a function instance by its name """
        abi_dict = self.get_abi(func_name)
        if not abi_dict:
            return None

        f = abi.Function(abi_dict)
        return f

    def get_events(self) -> List[abi.Event]:
        """ Get events from the abi sections """
        return [
            abi.Event(each) for each in self.get_abis() if each.get("type") == "event"
        ]

    def get_event_by_signature(self, signature: bytes) -> Union[abi.Event, None]:
        """
        Get target event from contract meta by signatuer string.

        Parameters
        ----------
        signature : bytes
            32 bytes
        """
        events_obj_list = self.get_events()
        events_dict = {each.get_signature(): each for each in events_obj_list}
        if signature in events_dict:
            return events_dict[signature]
        else:
            return None


import setuptools

long_description = ""
with open("README.md", "r") as fh:
    long_description = fh.read()
assert long_description

setuptools.setup(
    name="thor-requests",
    version="0.1.3",
    author="laalaguer",
    author_email="laalaguer@gmail.com",
    description="VeChain for humans",
    keywords="vechain thor blockchain sdk requests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/laalaguer/thor-requests",
    project_urls={
        "Documentation": "https://github.com/laalaguer/thor-requests",
        "Source": "https://github.com/laalaguer/thor-requests",
        "Issue Tracker": "https://github.com/laalaguer/thor-requests/issues",
    },
    python_requires=">=3.6",
    install_requires=[x.strip() for x in open("requirements.txt")],
    packages=setuptools.find_packages(),
)


import time
from typing import Union, List
import requests
from .utils import (
    build_tx_body,
    build_url,
    calc_blockRef,
    calc_chaintag,
    calc_emulate_tx_body,
    calc_nonce,
    any_emulate_failed,
    calc_revertReason,
    calc_tx_signed,
    calc_tx_unsigned,
    is_readonly,
    read_vm_gases,
    calc_gas,
    build_params,
)
from .wallet import Wallet
from .contract import Contract


class Connect:
    """ Connect to VeChain """

    def __init__(self, url):
        self.url = url

    def get_account(self, address: str, block: str = "best") -> dict:
        """ Query account status against the "best" block (or your choice)"""
        url = build_url(self.url, f"/accounts/{address}?revision={block}")
        r = requests.get(url, headers={"accept": "application/json"})
        if not (r.status_code == 200):
            raise Exception(f"Cant connect to {url}, error {r.text}")
        return r.json()

    def get_block(self, id_or_number: str = "best") -> dict:
        """ Get a block by id or number, default get "best" block """
        url = build_url(self.url, f"blocks/{id_or_number}")
        r = requests.get(url, headers={"accept": "application/json"})
        if not (r.status_code == 200):
            raise Exception(f"Cant connect to {url}, error {r.text}")
        return r.json()

    def get_chainTag(self) -> int:
        """ Fetch ChainTag from the remote network """
        b = self.get_block(0)
        return calc_chaintag(b["id"][-2:])

    def get_tx(self, tx_id: str) -> Union[dict, None]:
        """ Fetch a transaction, if not found then None """
        url = build_url(self.url, f"/transactions/{tx_id}")
        r = requests.get(url, headers={"accept": "application/json"})
        if not (r.status_code == 200):
            raise Exception(f"Cant connect to {url}, error {r.text}")
        return r.json()

    def post_tx(self, raw: str) -> dict:
        """
        Post tx to remote node. Get response.

        Parameters
        ----------
        raw : str
            '0x....' raw tx

        Returns
        -------
        dict
            post response eg. {'id': '0x....'}

        Raises
        ------
        Exception
            http exception
        """
        url = build_url(self.url, "transactions")
        r = requests.post(
            url,
            headers={"accept": "application/json", "Content-Type": "application/json"},
            json={"raw": raw},
        )
        if not (r.status_code == 200):
            raise Exception(f"Creation error? HTTP: {r.status_code} {r.text}")

        return r.json()

    def get_tx_receipt(self, tx_id: str) -> Union[dict, None]:
        """ Fetch tx receipt as a dict, or None """
        url = build_url(self.url, f"transactions/{tx_id}/receipt")
        r = requests.get(url, headers={"accept": "application/json"})
        if not (r.status_code == 200):
            raise Exception(f"Creation error? HTTP: {r.status_code} {r.text}")

        return r.json()

    def wait_for_tx_receipt(self, tx_id: str, timeout: int = 20) -> dict:
        """
        Wait for tx receipt, for several seconds

        Parameters
        ----------
        tx_id : str
            tx id
        timeout : int, optional
            seconds, by default 20

        Returns
        -------
        dict
            The receipt or None
        """
        interval = 3
        rounds = timeout // interval
        receipt = None
        for _ in range(rounds):
            receipt = self.get_tx_receipt(tx_id)
            if receipt:
                return receipt
            else:
                time.sleep(3)
        return None

    def emulate(self, emulate_tx_body: dict, block: str = "best") -> List[dict]:
        """
        Helper function. Use with caution.
        Upload a tx body for emulation,
        Get a list of execution responses (as the tx has multiple clauses).
        The response json structure please view README.md

        Parameters
        ----------
        emulate_tx_body : dict
            Emulate Tx body, not a normal tx body.
        block : str, optional
            Target at which block, by default "best"

        Returns
        -------
        List[dict]
            A list of clause execution results. (within the tx)

        Raises
        ------
        Exception
            If http has error.
        """
        url = build_url(self.url, f"/accounts/*?revision={block}")
        r = requests.post(
            url,
            headers={"accept": "application/json", "Content-Type": "application/json"},
            json=emulate_tx_body,
        )
        if not (r.status_code == 200):
            raise Exception(f"HTTP error: {r.status_code} {r.text}")

        all_responses = r.json()
        # decode the "revert" reason if emulation failed
        for response in all_responses:
            if response["reverted"] == True and response["data"] != "0x":
                response["decoded"] = {
                    "revertReason": calc_revertReason(response["data"])
                }
        return all_responses

    def replay_tx(self, tx_id: str) -> List[dict]:
        """
        Use the emulate function to replay the tx softly (for debug)
        Usually when you replay the tx to see what's wrong.

        Parameters
        ----------
        tx_id : str
            Existing tx id

        Returns
        -------
        List[dict]
            A list of clause execution results. (within the tx)

        Raises
        ------
        Exception
            If tx id doesn't exist
        """
        tx = self.get_tx(tx_id)
        if not tx:
            raise Exception(f"tx: {tx_id} not found")

        caller = tx["origin"]
        target_block = tx["meta"]["blockID"]
        emulate_body = calc_emulate_tx_body(caller, tx)
        if tx["delegator"]:
            emulate_body["gasPayer"] = tx["delegator"]

        return self.emulate(emulate_body, target_block)

    def emulate_tx(self, address: str, tx_body: dict, block: str = "best"):
        """
        Use the emulate function to emulate execution of a transaction.
        """
        emulate_body = calc_emulate_tx_body(address, tx_body)
        return self.emulate(emulate_body, block)

    def call(
        self,
        caller: str,
        contract: Contract,
        func_name: str,
        func_params: List,
        to: str,
        value=0,
        gas=0,  # Note: value is in Wei
        strict_mode=False,
    ) -> dict:
        """
        Call a contract method (read-only).
        This is a single transaction, single clause call.
        This WON'T create ANY change on blockchain.
        Only emulation happens.

        Response type view README.md
        If function has any return value, it will be included in "decoded" field
        """
        abi_dict = contract.get_abi(func_name)
        if not abi_dict:
            raise Exception(f"Function {func_name} not found on the contract")

        # In strict mode you can only call "read-only" functions.
        # Non-strict mode you can emulate "state changing" functions.
        if strict_mode:
            if not is_readonly(abi_dict):
                raise Exception(
                    f"Function {func_name} is not read-only, it is {abi_dict['stateMutability']}"
                )

        f = contract.get_function_by_name(func_name)
        data = f.encode(func_params, to_hex=True)  # Tx clause data
        clause = {"to": to, "value": str(value), "data": data}
        tx_body = build_tx_body(
            [clause],
            self.get_chainTag(),
            calc_blockRef(self.get_block("best")["id"]),
            calc_nonce(),
            gas=gas,
        )

        e_responses = self.emulate_tx(caller, tx_body)
        if len(e_responses) > 1:
            raise Exception(
                f"Emulation should contain 1 clause but has {len(e_responses)}"
            )
        failed = any_emulate_failed(e_responses)
        if failed:
            return e_responses[0]
        else:
            first_clause = e_responses[0]
            # decode the "return data" from the function call
            if first_clause["data"] and first_clause["data"] != "0x":
                first_clause["decoded"] = f.decode(
                    bytes.fromhex(first_clause["data"][2:])  # Remove '0x'
                )
            # decode the "event" from the function call
            if len(first_clause["events"]):
                for each_event in first_clause["events"]:
                    if each_event["address"].lower() == to.lower():
                        e_obj = contract.get_event_by_signature(
                            bytes.fromhex(each_event["topics"][0][2:])
                        )
                        if e_obj:
                            each_event["decoded"] = e_obj.decode(
                                bytes.fromhex(each_event["data"][2:]),
                                [bytes.fromhex(x[2:]) for x in each_event["topics"]],
                            )
                            each_event["name"] = e_obj.get_name()
            return first_clause

    def commit(
        self,
        wallet: Wallet,
        contract: Contract,
        func_name: str,
        func_params: List,
        to: str,
        value=0,  # Note: value is in Wei
        gas=0,
    ) -> dict:
        """
        Call a contract method,
        Similar to "call()" but will create state change to blockchain.
        And will spend gas.

        Parameters
        ----------
        wallet : Wallet
            Function Caller wallet
        contract : Contract
            Smart contract meta
        func_name : str
            Function name
        func_params: list
            Function params. eg. ['0x123..efg', '100']
        value:
            VET in Wei to send with this call
        gas:
            Gas you willing to pay to power this contract call.

        Returns
        -------
            Return value see post_tx()
        """
        abi_dict = contract.get_abi(func_name)
        if not abi_dict:
            raise Exception(f"Function {func_name} not found on the contract")

        if is_readonly(abi_dict):
            raise Exception(
                f"Function {func_name} is {abi_dict['stateMutability']}, use call() not commit()"
            )

        f = contract.get_function_by_name(func_name)
        data = f.encode(func_params, to_hex=True)  # Tx clause data
        clause = {"to": to, "value": str(value), "data": data}
        tx_body = build_tx_body(
            [clause],
            self.get_chainTag(),
            calc_blockRef(self.get_block("best")["id"]),
            calc_nonce(),
            gas=gas,
        )

        # We emulate it first.
        e_responses = self.emulate_tx(wallet.getAddress(), tx_body)
        if any_emulate_failed(e_responses):
            raise Exception(f"Tx will revert: {e_responses}")

        # Get gas estimation from remote
        _vm_gases = read_vm_gases(e_responses)
        _supposed_vm_gas = _vm_gases[0]
        _tx_obj = calc_tx_unsigned(tx_body)
        _intrincis_gas = _tx_obj.get_intrinsic_gas()
        _supposed_safe_gas = calc_gas(_supposed_vm_gas, _intrincis_gas)
        if gas and gas < _supposed_safe_gas:
            raise Exception(f"gas {gas} < emulated gas {_supposed_safe_gas}")

        # Fill out the gas for user.
        if not gas:
            tx_body["gas"] = _supposed_safe_gas

        encoded_raw = calc_tx_signed(wallet, tx_body, True)
        return self.post_tx(encoded_raw)

    def deploy(
        self,
        wallet: Wallet,
        contract: Contract,
        params_types: list = None,  # Constructor params types
        params: list = None,  # Constructor params
        value=0,  # send VET in Wei with constructor call
    ) -> dict:
        """
        Deploy a smart contract to blockchain
        This would be a single transaction.

        Parameters
        ----------
        wallet : Wallet
            Deployer wallet
        contract : Contract
            Smart contract meta
        params_types : list
            Constructor call parameter types. eg. ['address', 'uint256']
        params: list
            constructor call params. eg. ['0x123..efg', '100']
        value:
            VET in Wei to send with deploy call

        Returns
        -------
            Return value see post_tx()
        """
        # Build the constructor call data.
        if not params_types:
            data_bytes = contract.get_bytecode()
        else:
            data_bytes = contract.get_bytecode() + build_params(params_types, params)
        data = "0x" + data_bytes.hex()

        # Build the tx body.
        clause = {"to": None, "value": str(value), "data": data}
        tx_body = build_tx_body(
            [clause],
            self.get_chainTag(),
            calc_blockRef(self.get_block("best")["id"]),
            calc_nonce(),
            gas=0,  # We will estimate the gas later
        )

        # We emulate it first.
        e_responses = self.emulate_tx(wallet.getAddress(), tx_body)
        if any_emulate_failed(e_responses):
            raise Exception(f"Tx will revert: {e_responses}")

        # Get gas estimation from remote
        _vm_gases = read_vm_gases(e_responses)
        _supposed_vm_gas = _vm_gases[0]
        _tx_obj = calc_tx_unsigned(tx_body)
        _intrincis_gas = _tx_obj.get_intrinsic_gas()
        _supposed_safe_gas = calc_gas(_supposed_vm_gas, _intrincis_gas)

        # Fill out the gas for user.
        tx_body["gas"] = _supposed_safe_gas

        encoded_raw = calc_tx_signed(wallet, tx_body, True)
        return self.post_tx(encoded_raw)